/*
 * (c) 2010-2014, SWD Embedded Systems Limited, http://www.kpda.ru
 */

// Module Description:  hid class driver

#include "egalax.h"
#include <gulliver.h>
#include <sys/usbdi.h>
#include <sys/io-hid.h>
#include <math.h>
#include <stdarg.h>
#include <sys/slog.h>
#include <sys/slogcodes.h>
#include <sys/hidut.h>
#include <sys/neutrino.h>
#include <sys/netmgr.h>

#define RESEND_PULSE                              _PULSE_CODE_MINAVAIL
#define RESTART_PULSE                             _PULSE_CODE_MINAVAIL+1
#define CALIBRATION_COMPLETED                     _PULSE_CODE_MINAVAIL+2
#define NEXT_RAW_DATA_PULSE                       _PULSE_CODE_MINAVAIL+5

#define WAIT_USB              0
#define USB_SERVER            1
#define VENDOR_ID             2
#define DEVICE_ID             3 
#define VERBOSE               4
#define NO_INIT               5
#define GET_CONTROLLER_INFO   6
#define HID                   7
#define SENSITIVITY           8
#define DELAY                 9

typedef struct raw_data {
_uint16 x,y;
} raw_data_t;

int usb_init( void *dll_hdl, dispatch_t *dpp, io_hid_self_t *ioh, char *options );
int usb_stop();

int usb_client_attach		(int reg_hdl, void *user);
int usb_client_detach		(int reg_hdl, void *user);
int usb_report_read			(int reg_hdl, void *user, _uint8 rid, _uint16 rtype, void *rdata, _uint16 rlen);
int usb_report_write		(int reg_hdl, void *user, _uint8 rid, _uint16 rtype, void *rdata, _uint16 rlen);
int usb_get_idle				(int reg_hdl, void *user, _uint8 rid, _uint16 *idle_rate);
int usb_set_idle				(int reg_hdl, void *user, _uint8 rid, _uint16 idle_rate);
int usb_get_protocol		(int reg_hdl, void *user, _uint8 *protocol);
int usb_set_protocol		(int reg_hdl, void *user, _uint8 protocol);
int usb_reset						(int reg_hdl, void *user);
int usb_rbuffer_alloc		(_uint16 size, void **bptr);
int usb_get_string			(int reg_hdl, void *user, _uint16 request, void **str);
int usb_get_indexed_string	(int reg_hdl, void *user, _uint16 index, void **str);
int usb_rbuffer_free		(void *bptr);
int convert_to_hid      (uint8_t *p, _uint32 alen);

int 					NumCtrl = 0;
usb_ctrl_t				USBCtrl;

struct hiddi_device_instance;

io_hid_dll_entry_t io_hid_dll_entry = {
	"devh-egalax",
	2,
	usb_init,
	usb_stop
};


io_hid_registrant_funcs_t usb_funcs = {
	9,
	usb_client_attach,
	usb_client_detach,
	usb_rbuffer_alloc,
	usb_rbuffer_free,
	usb_report_read,
	usb_report_write,
	usb_get_idle,
	usb_set_idle,
	usb_get_protocol,
	usb_set_protocol,
	usb_get_string,
	usb_get_indexed_string,
	usb_reset,
};

char *usb_opts[] = {
	"wait",
	"upath",
	"vid",
	"did",
	"verbose",
	"noinit",
	"info",
	"hid",
	"sensitivity",
	"delay",
	NULL
};

static unsigned char touchscreen_rdesc[] = {
0x05, 0x01,     // Usage Page(Generic Desktop)
0x09, 0x01,     // Usage(Pointer)
0xa1, 0x01,     // Collection(Application)
0x09, 0x01,         // Usage(Pointer)
0xa1, 0x00,         // Collection(Physical)
0x05, 0x09,             // Usage Page(Button)
0x19, 0x01,             // Usage Min(1)
0x29, 0x01,             // Usage Max(1)
0x15, 0x00,             // Logical Min(0)
0x25, 0x01,             // Logical Max(1)
0x75, 0x01,             // Report Size(1)
0x95, 0x01,             // Report Count(1)
0x81, 0x02,             // Input(Data,Variable,Absolute)  <-
0x75, 0x07,             // Report Size(7)                 <-
0x81, 0x01,             // Input(Constant,Array,Absolute)
0x05, 0x01,             // Usage Page(Generic Desktop)
0x09, 0x30,             // Usage(X)
0x09, 0x31,             // Usage(Y)
0x09, 0x32,             // Usage(Z)
0x75, 0x10,             // Report Size(16)
0x95, 0x03,             // Report Count(3)
0x81, 0x62,             // Input(Data,Variable,Absolute)
0x05, 0x00,  						// Usage Page (Undefined)
0x19, 0x01,             // Usage Minimum (1)
0x29, 0xFF,             // Usage Maximum (255)
0x95, 0x08,             // Report Count (8)
0x75, 0x08,             // Report Size (8)
0x91, 0x02,             // Output (Data, Variable, Absolute)
0xc0,               // End Collection
0xc0            // End Collection
};


/******************************************************************************

  Slog wrapper function

******************************************************************************/
ssize_t hid_slogf (int opcode, int severity, int level, const char *fmt, ...)
{
	ssize_t ret;
	va_list	arglist;
	int verbosity;

	ret = 0;
	verbosity = USBCtrl.verbose;

	if (verbosity >= level) {
		va_start (arglist, fmt);
		ret = vslogf (opcode, severity, fmt, arglist);
		va_end (arglist);
	}

	return (ret);
}


/******************************************************************************

	Checks for the stall condition.

******************************************************************************/
_uint32 usb_xlat_error( _uint32 status )
{
	switch (status & USBD_USB_STATUS_MASK) {	
	case USBD_STATUS_STALL :
		return (ENOTSUP);
	}

	return (EIO);
}


/******************************************************************************

	Clear the device in use status and close the endpoints that are in use.

******************************************************************************/
int usb_free	(usb_t *usb)
{
	int status;

	usb->status	&= ~USB_DEVICE_IN_USE;

	// Stop the Interrupt In endpoint
	if (usb->ep_int_in) {
		while (usbd_close_pipe (usb->ep_int_in) == EBUSY) {
			usbd_abort_pipe (usb->ep_int_in);
			delay (10);
		}
	}

	// Stop the Interrupt Out endpoint
	if (usb->ep_int_out) {
		while (usbd_close_pipe (usb->ep_int_out) == EBUSY) {
			usbd_abort_pipe (usb->ep_int_out);
			delay(10);
		}
	}

	// Stop the Control endpoint
	if (usb->ep_cntl) {
		while (usbd_close_pipe (usb->ep_cntl) == EBUSY) {
			usbd_abort_pipe (usb->ep_cntl);
			delay (10);
		}
	}

	if (usb->urb)
		usbd_free_urb (usb->urb);

	usbd_free (usb->rbuffer);

	TAILQ_REMOVE (&USBCtrl.dlist, usb, dlink);

	if ((status = usbd_detach (usb->device)) == EBUSY) {
		hid_slogf (_SLOGC_USB, 2, 0, "devh-egalax.so - usbd_detach %x", status);
	}

	return (status);
}


/******************************************************************************

	Signal handler

******************************************************************************/
void usb_signal_handler (int signo)
{
	usb_t *usb;
	usb_t *nusb;
	int x;	

	signo = signo;

	for (usb = TAILQ_FIRST (&USBCtrl.dlist); usb; usb = nusb) {
		nusb = TAILQ_NEXT (usb, dlink);
		usb_free (usb);
	}

	for (x = 0 ; x < USBCtrl.num_conn ; x++)
		usbd_disconnect (USBCtrl.connection[x]);
}


/******************************************************************************

	Wrapper function to send a vendor specific control command to the device.

******************************************************************************/
_uint32 usb_send_control_command (usb_t *usb, _uint32 dir, _uint16 command, _uint16 request_type, _uint16 parameter, void *data, int dlen) 
{
	_uint32 status = ENOMEM;
	struct usbd_urb *urb;
	_uint8 *d = NULL;
	_uint32 len;


	if (USBCtrl.verbose > 4)
		fprintf (stderr, "Request type %x, Command:%x, Parameter %x, Direction %x\n", request_type, command, parameter, dir);

	if (!dlen || (d = (_uint8 *)usbd_alloc (dlen))) {
		memcpy (d, data, dlen);

		if ((urb = usbd_alloc_urb (NULL))) {
			usbd_setup_vendor (urb, dir, command, request_type, parameter, usb->usbd_instance.iface, (void *)d, dlen);

			if ((status = usbd_io (urb, usb->ep_cntl, NULL, usb, USBD_TIME_INFINITY)) == EOK || status == EIO) {
				if (usbd_urb_status (urb, &status, &len) == EOK  && !(status & USBD_STATUS_CMP_ERR)) {
					memcpy( data, d, min(len, dlen) );
					status = EOK;
				}
				else {
					status = usb_xlat_error (status);
				}
			}
			else {
				status = EIO;
			}

			usbd_free_urb (urb);
		}

		if (dlen)
			usbd_free (d);
	}

	return (status);
}



/******************************************************************************

  Send a simple loopback command to the controller.

******************************************************************************/
int egalax_send_loopback (usb_t *usb)
{
	int status = EOK;
	_uint8 buf[3];

	memset (buf, 0, sizeof (buf));

	/* Check to see if the controller is active */
	buf[0] = HEADER;
	buf[1] = CMD_LENGTH;
	buf[2] = 0x41;

	if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, buf, sizeof (buf))) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to send check active (loopback) command to the controller: %d", status);
	}

	return (status);
}


/******************************************************************************

  Send a reset command to the controller.

******************************************************************************/
int egalax_send_reset (usb_t *usb)
{
	int status = EOK;
	_uint8 buf[3];

	memset (buf, 0, sizeof (buf));

	/* Issue a reset to the controller */
	buf[0] = HEADER;
	buf[1] = CMD_LENGTH;
	buf[2] = 0x67;

	if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, buf, sizeof (buf))) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to send reset command to the controller: %d", status);
	}

	return (status);
}


/******************************************************************************

  Set the sensitivity for the touchscreen (80 - 255) on the controller.

******************************************************************************/
int egalax_set_sensitivity (usb_t *usb)
{
	int status = EOK;
	_uint8 buf[6];

	memset (buf, 0, sizeof (buf));

	/* Shouldn't go lower than 80 */
	if (USBCtrl.sensitivity < 80)
		return (-1);

	/* Set the sensitivity of the controller */
	buf[0] = HEADER;
	buf[1] = 0x04;
	buf[2] = SET_SENSITIVITY;
	buf[3] = 0x2A;
	buf[4] = USBCtrl.sensitivity;
	buf[5] = 0x00;

	if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, buf, sizeof (buf))) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to send the set sensitivity command to the controller: %d", status);
	}

	egalax_send_reset (usb);

	return (status);
}


/******************************************************************************

  Request the firmware version on the controller.

******************************************************************************/
int egalax_request_firmware_revision (usb_t *usb)
{
	int status = EOK;
	_uint8 buf[3];

	memset (buf, 0, sizeof (buf));

	/* Issue a request for the firmware version */
	buf[0] = HEADER;
	buf[1] = CMD_LENGTH;
	buf[2] = REQUEST_FIRMWARE_REVISION;

	if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, buf, sizeof (buf))) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to send request firmware command to the controller: %d", status);
	}

	return (status);
}


/******************************************************************************

  Set the delay for the touchscreen (0 - 2550us) on the controller.

******************************************************************************/
int egalax_set_delay (usb_t *usb)
{
	int status = EOK;
	_uint8 buf[6];

	memset (buf, 0, sizeof (buf));
#if 0
	/* We don't know what this command is yet, need to figure it out for later */

	/* Shouldn't go lower than 80 */
	if (USBCtrl.delay > 2550)
		return (-1);

	/* Check to see if the controller is active */
	buf[0] = HEADER;
	buf[1] = 0x04;
	buf[2] = 0x42;
	buf[3] = 0x37;
	buf[4] = USBCtrl.touch_delay;
	buf[5] = 0x00;

	if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, buf, 3)) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to send set delay command to the controller: %d", status);
	}

	egalax_send_reset (usb);
#endif
	return (status);
}

/******************************************************************************

	Reset the controller using vendor specific reset and init commands. 

******************************************************************************/
int usb_device_reset (usb_t *usb)
{
	int	status = EOK;
	_uint8 buf[3];

	memset (buf, 0, sizeof (buf));

	if (!(USBCtrl.cflags & FLAG_NO_INIT)) {
		if ((status = usbd_reset_device (usb->device)) == EOK) {
			egalax_send_reset (usb);
		}
	}

	return (status);
}


/******************************************************************************

  Once all the raw data points have been sent to the controller for calibration
  we send this command to tell the firmware to use those points to calculate
  the calibration.

******************************************************************************/
int egalax_calculate_calib_params (usb_t *usb)
{
	int status = EOK;
	_uint8 buf[4];

	memset (buf, 0, sizeof (buf));

	buf[0] = HEADER;
	buf[1] = 0x02;
	buf[2] = 0x37;//USE_RAW_COORDINATES;
	buf[3] = 0xFF;

	if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, buf, sizeof (buf))) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to send calculate calibration parameters command to the controller: %d", status);
	}

	return (status);
}


/******************************************************************************

  Switch the controllers operating mode between RAW and Calibrated

******************************************************************************/
int egalax_set_mode (usb_t *usb, int cmd)
{
  int status = EOK;
  _uint8 buf[4];

	memset (buf, 0, sizeof (buf));

  buf[0] = HEADER;
  buf[1] = 0x02;
  buf[2] = SET_REPORT_MODE;
  buf[3] = cmd;

  if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, buf, sizeof (buf))) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to set the controller mode: %d", status);
  } else {
		hid_slogf (_SLOGC_USB, 1, 0, "devh-egalax.so - Controller set to %s", cmd ? "Calibrated point" : "Raw");
	}

	if (USBCtrl.verbose)
		fprintf (stderr, "Set mode to %s!\n", cmd ? "Calibrated point" : "Raw");

  return (status);
}


/******************************************************************************

  Send the raw calibration data points to the controller.

******************************************************************************/
int egalax_send_calib_data (usb_t *usb, egalax_data_packet_t *data) 
{
	int status = EOK;
	_uint8 cmd[6];
	_uint8 buf[3];

	memset (cmd, 0, sizeof (cmd));
	memset (buf, 0, sizeof (buf));

	/* Send Raw calibrated points to the controller */
	cmd[0] = HEADER;
	cmd[1] = 0x07;
	cmd[2] = 0x37;

	/* Set the calibration type for the firmware */
	switch (data->num_points) {
	case 4:
		cmd[3] = 0x00;
		break;
	case 9:
		cmd[3] = 0x01;
		break;
	case 25:
		cmd[3] = 0x02;
		break;
	default:
		cmd[3] = 0x02;
		break;
	}

	cmd[4] = data->index;                 /* Index for the firmware */
	cmd[5] = data->x & 0xFF;              /* X Low Byte */

	/* We need to do the command in two stages as the FIFO on the IC is only 8 bytes, this is not doc'd */
	if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, cmd, sizeof (cmd))) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to send the controller the first set of raw data points: %d", status);
		return (-1);
	}

	if (USBCtrl.verbose)
		fprintf (stderr, "Writing data points %d X%d Y%d to the EEPROM\n", data->index, data->x, data->y);

	buf[0] = (data->x>> 8) & 0xFF;        /* X High Byte */
	buf[1] = data->y & 0xFF;              /* Y Low Byte */
	buf[2] = (data->y >> 8) & 0xFF;       /* Y High Byte */

	if ((status = usb_send_control_command (usb, URB_DIR_OUT, 0, USB_TYPE_VENDOR, 0, buf, sizeof (buf))) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Failed to send the controller the second set of raw data points: %d", status);
		return (-1);
	}

	return (0);
}


/******************************************************************************

******************************************************************************/
int process_reply (usb_t *usb, _uint8 *reply)
{
	switch (reply[2]) {
		case CHECK_ACTIVE:
			hid_slogf (_SLOGC_USB, 1, 0, "devh-egalax.so - Egalax touch screen has been initalized and is active.");
			break;
		case REQUEST_FIRMWARE_REVISION:
		{
			char *firmware_revision;

			/* Jump to the returned string */
			firmware_revision = (char *)(reply + 3);
			hid_slogf (_SLOGC_USB, 1, 0, "devh-egalax.so - Firmware revision %s", firmware_revision);
			break;
		}
		case SET_SENSITIVITY_REPLY:
			break;
		case REQUEST_CONTROLLER_TYPE:
		{
			char *controller_type;

			/* Jump to the returned string */
			controller_type = (char *)(reply + 3);
			hid_slogf (_SLOGC_USB, 1, 0, "devh-egalax.so - Controller Type %s", controller_type);
			break;
		}
		case SET_REPORT_MODE_ACK:
			hid_slogf (_SLOGC_USB, 1, 0, "devh-egalax.so - Controller set to %s", reply[3] ? "Calibrated point" : "Raw");
			break;
		case RESET_CONTROLLER:
		{
			struct sched_param param;
			int   sched_policy;
			int   priority;

			hid_slogf (_SLOGC_USB, 1, 0, "devh-egalax.so - Controller reset was successful");

			/* If we are in the middle of sending coordinat infomation from calib we should alert calib that the controller reset */
			if (USBCtrl.sending_calibration_data) {
				pthread_getschedparam (pthread_self(), &sched_policy, &param);
				priority = param.sched_curpriority + 2;

				MsgSendPulse (usb->coid, priority, RESTART_PULSE, 2);
			}
		}
		break;
		case CALIBRATION_PARAMETERS:
		{
			int limit = 0;
			int priority = 0;
			struct sched_param param;
			int   sched_policy;

			if (USBCtrl.verbose)
				fprintf (stderr, "Received a reply to the last data position for calibration\n");

			/* Check to see if it's a response to calculate calibration params */
			if (reply[3] == 0xFF) {
				hid_slogf (_SLOGC_USB, 1, 0, "devh-egalax.so - Egalax touch screen linearization has completed sucessfully.");
				break;
			} else if (usb->coid == NULL) {
				/* Got this for some unknown reason, bail out */
				break;
			}

			switch (reply[3]) {
			case 0x00:        /* 4 Point Calibration */
				limit = 3;
				break;
			case 0x01:        /* 9 Point Calibration */
				limit = 8;
				break;
			case 0x02:        /* 25 Point Calibration */
				limit = 24;
				break;
			default:
				break;
			}

			if (USBCtrl.verbose > 5)
				fprintf (stderr, "We have Sent %d packets\n", reply[4]);

			/* Check to see if that was the last entry to send */
			if (limit == reply[4]) {
				/* We are all done disconnect */
				name_close (usb->coid);
				usb->coid = NULL;
			} else {
				pthread_getschedparam (pthread_self(), &sched_policy, &param);
				priority = param.sched_curpriority + 2;

			 	if (MsgSendPulse (usb->coid, priority, NEXT_RAW_DATA_PULSE, 0) == -1) {
					hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - MsgSendPulse failed!: %d", errno);
					break;
				}
			} 
		}
		break;
		default:
			/* Check to see if this is a actually a touch point */
			hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - Received an unknown response to a command: %d", reply[2]);
			break;
	}

	return (0);
}


/******************************************************************************

  See if there is a full reply in the ring buffer if so remove and process it.

******************************************************************************/
int reply_handler (usb_t *usb)
{
	int tmp_start = 0;
	_uint8 p[256];
	int i;

	memset (p, 0, sizeof (p));

	tmp_start = usb->ring_start;

	if (USBCtrl.verbose > 6) {
		/* Dump ring buffer contents */
		fprintf (stderr, "Ring buffer start pos %d, end pos: %d, contents: ", usb->ring_start, usb->ring_end);
		while (tmp_start != usb->ring_end) {
			fprintf (stderr, "%x ", usb->ring_buffer[tmp_start]);
			tmp_start = (tmp_start + 1) % RING_BUFFER_SIZE;
		}
		fprintf (stderr, "\n");
		tmp_start = usb->ring_start;
	}

	while (usb->ring_start != usb->ring_end) {
		/* Check the contents of the buffer to see if we have a full reply on it */
		if (usb->ring_buffer[tmp_start] == HEADER) {
			tmp_start = (tmp_start + 1) % RING_BUFFER_SIZE;

			/* Firmware reply, lets check the length and make sure its all there */
			if (usb->ring_buffer[tmp_start] <= usb->ring_buffer_current_size) {
				/* We have enough, remove contents from ring buffer and process it. */
				for (i = 0; i <= (usb->ring_buffer[tmp_start] + 1); i++) {
					p[i] = usb->ring_buffer[usb->ring_start];
					usb->ring_start = (usb->ring_start + 1) % RING_BUFFER_SIZE;
				}

				usb->ring_buffer_current_size -= i;
				process_reply (usb, p);
			} else {
				/* Still waiting on the rest of the reply */
				break;
			}
		} else if (usb->ring_buffer[tmp_start] & SYNC_BIT) {
			int touch_packet_len = 5;

			/* Touch packet, lets check the length, touch packets can be between 5 and 6 bytes in length */
			if ((usb->ring_buffer[tmp_start] & PRESSURE_SET) || (usb->ring_buffer[tmp_start] & PLAYER_ID_SET))
				touch_packet_len = 6;

			if (touch_packet_len <= usb->ring_buffer_current_size) {
				/* Remove the packet from the buffer */
				for (i = 0; i < touch_packet_len; i++) {
					p[i] =  usb->ring_buffer[usb->ring_start];
					usb->ring_start = (usb->ring_start + 1) % RING_BUFFER_SIZE;
				}

				usb->ring_buffer_current_size -= touch_packet_len;

				/* Convert it to a HID compliant packet */
 	     if (convert_to_hid (p, touch_packet_len)) {
 	       if ((*USBCtrl.ioh->get_buffer) (usb->hid_hdl, &usb->rbuffer) != EOK) {
 	       }
        /* Copy the buffer's content into the report buffer */
        memcpy (usb->rbuffer, p, touch_packet_len);

        (*USBCtrl.ioh->send_report) (usb->hid_hdl, (void *) usb->rbuffer, (_uint32) touch_packet_len);
 	     }
			} else {
				/* Still waiting on the rest of the packet */
				break;
			}
		} else {
			/* Not sure what this is */
			hid_slogf (_SLOGC_USB, 2, 0, "devh-egalax.so - Note: Ring buffer has an unknown packet in the start %x these are often junk, skipping.", usb->ring_buffer[tmp_start]);
			usb->ring_start = (usb->ring_start + 1) % RING_BUFFER_SIZE;
			usb->ring_buffer_current_size--;
		}
	}

	return (0);
}


/******************************************************************************

	Convert a data packet to HID (See HID Spec)

	HID Expects:
	p[0] = Buttons 0x1 = touch
                 0x0 = release
  p[1] = Lower X
  p[2] = Upper X
  p[3] = Lower Y
  p[4] = Upper Y

******************************************************************************/
int convert_to_hid (uint8_t *p, _uint32 alen)
{
	int i;
	unsigned x = 0, y = 0, z = 0;
	int max_val = 0, bit_resolution = 0;
	int data_packet_length = 5;

	if (USBCtrl.verbose > 3)
		fprintf (stderr, "Length of data %d\n", alen);

	if (USBCtrl.verbose > 5) {
		for (i = 0; i < alen; i++)
			fprintf (stderr, "Coordinate Data is %x\n", *(p + i));
	}

	/* Convert to HID */
	for (i = 0; i < alen; i++) {
		/* Check to make sure we are in sync, sometimes we get controllers that like to send an extra byte first, not
		   sure what that is for yet, so for now, find the sync byte and go from there. */
		if ((p[i] & SYNC_BIT) && (i == 0)) {
			/* Check the matrix resolution */
			switch ((p[i] >> 1) & 0x03) {
			case 0:
				max_val = 2048;
				bit_resolution = 11;
				break;
			case 1:
				max_val = 4096;
				bit_resolution = 12;
				break;
			case 2:
				max_val = 8192;
				bit_resolution = 13;
				break;
			case 3:
				max_val = 16384;
				bit_resolution = 14;
				break;
			default:
				break;
			}

			if (USBCtrl.verbose > 3) 
				fprintf (stderr, "Touchscreen has %dbit resolution (max %d).\n", bit_resolution, max_val);

			/* Check to see if pressure information is going to be set */
			if (p[i] & PRESSURE_SET) {
				z = p[i + 5]; 

				if (USBCtrl.verbose > 3) 
					fprintf (stderr,"Pressure is set\n");

				data_packet_length = 6;
			}

			if (p[i] & PLAYER_ID_SET) {
				if (USBCtrl.verbose > 3) 
					fprintf (stderr, "Player ID is set\n");

				data_packet_length = 6;
			}

	    /* Convert from Egalax format to x/y data. */
			x = (p[i + 1] << 7) | p[i + 2]; 
			y = (p[i + 3] << 7) | p[i + 4];

			/* Check to see if we are in raw mode, if so then this limit might need to be ignored */
			if ((x > max_val) || (y > max_val)) {
				hid_slogf (_SLOGC_USB, 2, 0, "devh-egalax.so - Touch Value has exceeded %d bit limit X:%d Y:%d.  Packet Dropped.", bit_resolution, x, y);
				return (0);
			}

			/* Convert the coordinates to HID compatible format */
			/* Button information */
			p[0] = p[i];

			/* X */
			p[2] = ((x & 0xFF00) >> 8);
			p[1] = (x & 0x00FF);

			/* Y */
			p[4] = ((y & 0xFF00) >> 8);
			p[3] = (y & 0x00FF);

			if (USBCtrl.verbose)
				fprintf (stderr, "%s, X%d, Y%d\n", (*p & 0x01) ? "Touched" : "Released", ((*(p + 2)) << 8) | *(p + 1), ((*(p + 4)) << 8) | *(p + 3));

			return (alen);
		}
	}

	return (0);
}



/******************************************************************************

	Interrupt callback routine.  Handles incomming data from the controller.

******************************************************************************/
void usb_int_cbf (struct usbd_urb *urb, struct usbd_pipe *pipe, void *hdl)
{
	usb_t *usb;
	uint8_t *buf;
	_uint32 ustatus;
	_uint32 alen;
	int i;

	usb = (usb_t *)hdl;
	buf = usb->ep_int_buf;

	if (USBCtrl.verbose > 5)
		fprintf (stderr, "Interrupt occured\n");

	usbd_urb_status (urb, &ustatus, &alen);

	if ((ustatus & USBD_URB_STATUS_MASK) != USBD_STATUS_CMP) {
		// check for a stall condition and clear it
		if ((ustatus & USBD_USB_STATUS_MASK) == USBD_STATUS_STALL) {
			//can't tell the difference on UHCI if a real stall or device is not responding
			if (usbd_reset_pipe (pipe)) {
				egalax_send_loopback (usb);
			}
		}
		else {
			if (usb->status & USB_DEVICE_IN_USE) {
				egalax_send_loopback (usb);
			}
		}
	}

	if ((usb->status & USB_DEVICE_IN_USE) && !(ustatus & USBD_USB_STATUS_MASK)) {
		usbd_setup_interrupt (usb->urb, URB_DIR_IN | URB_SHORT_XFER_OK, usb->ep_int_buf, usb->ep_int_in_size);

		/* Stuff the values into the ring buffer and see if we have enough to process */
		for (i = 0; i < alen; i++) {
			if ((usb->ring_end + 1) % RING_BUFFER_SIZE != usb->ring_start) {
				usb->ring_buffer[usb->ring_end] = buf[i];
				usb->ring_end = (usb->ring_end + 1) % RING_BUFFER_SIZE;
			} else {
				hid_slogf (_SLOGC_USB, 2, 0, "devh-egalax.so - Ring Buffer is out of space, dropping packet.");
			}
		}

		usb->ring_buffer_current_size += alen;

		if (usbd_io (usb->urb, pipe, usb_int_cbf, usb, USBD_TIME_INFINITY)) {
			usb->status &= ~USB_DEVICE_IN_USE;
		}

		/* Process the reply */
		reply_handler (usb);

	}
	/* use the same buffer */
	else if ((usb->status & USB_DEVICE_IN_USE) && !( ustatus & USBD_STATUS_ABORTED)) {
		usbd_setup_interrupt (usb->urb, URB_DIR_IN | URB_SHORT_XFER_OK, usb->rbuffer, usb->ep_int_in_size);

		if (usbd_io (usb->urb, pipe, usb_int_cbf, usb, USBD_TIME_INFINITY)) {
			usb->status &= ~USB_DEVICE_IN_USE;
		}
	}
}


/******************************************************************************

	Setup report descriptor and register with io-hid.

******************************************************************************/
int usb_get_report_descriptor (usb_t *usb, int intnum , struct usbd_desc_node *ifc)
{
	void *report_data = NULL;
	int usb_hdl;
	usbd_device_descriptor_t *ddevice;
	hidd_device_ident_t device_ident;
	io_hid_registrant_t dev;
	int result = EOK;

	device_ident.vendor_id = usb->usbd_instance.ident.vendor;
	device_ident.product_id = usb->usbd_instance.ident.device;

	if ((ddevice = usbd_device_descriptor (usb->device, NULL)) == NULL) {
		usbd_free (report_data);
		return (EIO);
	}

	device_ident.version = ddevice->bcdUSB;
	dev.flags = 0;
	dev.device_ident = &device_ident;
	dev.desc = touchscreen_rdesc; 
	dev.dlen = sizeof (touchscreen_rdesc);
	dev.user_hdl = usb;
	dev.funcs = &usb_funcs;

	/* register device with hiddi */
	if ((result = (*USBCtrl.ioh->reg) (USBCtrl.dll_hdl, &dev, &usb_hdl)) == EOK) 
		usb->hid_hdl = usb_hdl;

	return (result);
}


/******************************************************************************

	Scan for the nessisary endpoints

******************************************************************************/
int usb_parse_descriptors (usb_t *usb)
{
	_uint32 eix;
	_uint32 scan;
	_uint32 found;
	usbd_descriptors_t *desc;
	struct usbd_desc_node *ifc, *ept;
	int status;

	scan = TOUCH_CONTROL_EP | TOUCH_INTIN_EP; 
	found	= 0;

	if (USBCtrl.verbose > 5)
		fprintf (stderr, "Parse Descriptors\n");

	// crete endpoints for all interfaces
	if ((desc = (usbd_descriptors_t *) usbd_interface_descriptor (usb->device, usb->usbd_instance.config, usb->usbd_instance.iface , usb->usbd_instance.alternate, &ifc))) {
		for (eix = 0; (desc = usbd_parse_descriptors (usb->device, ifc, USB_DESC_ENDPOINT, eix, &ept)) != NULL; ++eix) {
			switch (desc->endpoint.bmAttributes) {
			case USB_ATTRIB_CONTROL:
				if (usbd_open_pipe (usb->device, desc, &usb->ep_cntl) == EOK) {
					found |= TOUCH_CONTROL_EP;
				}
				break;

			case USB_ATTRIB_ISOCHRONOUS:
				break;

			case USB_ATTRIB_BULK:
				break;

			case USB_ATTRIB_INTERRUPT:
				switch (desc->endpoint.bEndpointAddress & USB_ENDPOINT_IN) {
				case USB_ENDPOINT_OUT:
					break;

				case USB_ENDPOINT_IN:
					if (usbd_open_pipe (usb->device, desc, &usb->ep_int_in) == EOK) {
						usb->ep_int_in_size = desc->endpoint.wMaxPacketSize;
						found |= TOUCH_INTIN_EP;

						if ((usb->urb = usbd_alloc_urb (NULL)) == NULL) {
							return (ENOMEM);
						}
					}
					break;
				}
				break;
			}
		}

		if ((desc = (usbd_descriptors_t *) usbd_interface_descriptor( usb->device, usb->usbd_instance.config, usb->usbd_instance.iface , usb->usbd_instance.alternate, &ifc))) {
			/* Since the Egalax controllers vary in their supported protocols and some report themselves as a HID compliant even when they are not, always attach to the
			   device we will check in the usb_device_reset function what controller we are using and determine wether or not to continue. */
			if ((status = usb_get_report_descriptor (usb, desc->interface.bInterfaceNumber, ifc)))
				return (status);
    }
	}

	return ((found == scan) ? EOK : ENODEV);
}


/******************************************************************************

	Insertion callback.

******************************************************************************/
void usb_insertion (struct usbd_connection *connection, usbd_device_instance_t *instance)
{
	usb_t *usb;
	struct usbd_device *device;
	int status;

	if (USBCtrl.verbose > 5)
		fprintf (stderr, "Device Inserted\n");

	if ((status = usbd_attach (connection, instance, sizeof (usb_t), &device)) == EOK) {
		usb = usbd_device_extra (device);
		usb->connection		= connection;
		usb->device			= device;
		usb->usbd_instance	= *instance;

		usb->idle_rate = 0xffff;
		usb->protocol = 0xffff;

		/* Start with a clean ring buffer */
		usb->ring_start = 0;
		usb->ring_end = 0;

		memset (usb->ring_buffer, 0, sizeof (usb->ring_buffer));


		TAILQ_INSERT_TAIL (&USBCtrl.dlist, usb, dlink);
		if (usb_parse_descriptors (usb) == EOK) {
			/* alloc space for buffers */
			if((usb->ep_int_buf = usbd_alloc(usb->ep_int_in_size))) {
			}
			return;	
		}

		usb_free (usb);
	}
	else {
		hid_slogf (_SLOGC_USB, 2, 0, "devh-egalax.so - Unable to attach to USB device %d (%x)",instance->devno, status);
	}
}


/******************************************************************************


******************************************************************************/
void *usb_dereg_thread (void *data)
{
	struct usbd_device	*device;
	usb_t				*usb;

	usb = (usb_t *) data;

	if ((device = usbd_device_lookup (usb->connection, &usb->usbd_instance)) != NULL) {
		if ( (usb = usbd_device_extra (device))) {
			usb->status	&= ~USB_DEVICE_IN_USE;
			(*USBCtrl.ioh->dereg) (usb->hid_hdl);
			usb_free (usb);
		}
	}

	return (NULL);
}


/******************************************************************************

	Removal callback, cleans up the driver.

******************************************************************************/
void usb_removal (struct usbd_connection *connection, usbd_device_instance_t *instance)
{
	struct usbd_device *device;
	usb_t *usb;
	pthread_attr_t attr;
	struct sched_param param;

	if (USBCtrl.verbose > 5)
		fprintf (stderr, "Device Removed\n");

	pthread_attr_init (&attr);
	pthread_attr_setschedpolicy (&attr, SCHED_RR);
	param.sched_priority = UMASS_DEREG_PRIORITY;
	pthread_attr_setschedparam (&attr, &param);
	pthread_attr_setinheritsched (&attr, PTHREAD_EXPLICIT_SCHED);
	pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize (&attr, 8192);

	if ((device = usbd_device_lookup (connection, instance)) != NULL) {
		if ((usb = usbd_device_extra (device))) {
			usb->status	&= ~USB_DEVICE_IN_USE;

			if (!pthread_create (NULL, &attr, (void *)usb_dereg_thread, usb)) {
			}
		}
	}
}


/******************************************************************************

	Client attaches.

	First attachment to the device, so enable the device.

******************************************************************************/
int usb_client_attach (int reg_hdl, void *user)
{
	usb_t		*usb = user;
	int			status;

	if (USBCtrl.verbose > 5)
		fprintf (stderr, "Client attached\n");
		
	if ((*USBCtrl.ioh->get_buffer) (usb->hid_hdl, &usb->rbuffer) != EOK) {
	}

	usb->status |= USB_DEVICE_IN_USE;
	usbd_setup_interrupt (usb->urb, URB_DIR_IN | URB_SHORT_XFER_OK, usb->rbuffer, usb->ep_int_in_size);

	if ((status = usbd_io (usb->urb, usb->ep_int_in, usb_int_cbf, usb, USBD_TIME_INFINITY))) {
		hid_slogf (_SLOGC_USB, 2, 0, "devh-egalax.so - error usbd_io %x", status);

		return (status);
	}

	/* Clear any old values just in case */
	USBCtrl.partial_reply_expected_len = 0;
	USBCtrl.partial_reply_current_len = 0;
	memset (USBCtrl.partial_reply, 0, sizeof (USBCtrl.partial_reply));

	/* Reset the device */
	usb_device_reset (usb);

	/* Request the firmware revision */
	egalax_request_firmware_revision (usb);

	/* Set the sensitivity of the controller */
/*
	if (USBCtrl.sensitivity)
		egalax_set_sensitivity (usb);
*/
	return (EOK);
}


/******************************************************************************

  Last client detached from the device, so disable the device. 

******************************************************************************/
int usb_client_detach (int reg_hdl, void *user)
{
	usb_t				*usb = user;
	
	usb->status	&= ~(USB_DEVICE_IN_USE);

	// abort devices pipes
	usbd_abort_pipe (usb->ep_int_in);
	
	return (EOK);
}


/******************************************************************************

  Client called read report (polls the device)

******************************************************************************/
int usb_report_read (int reg_hdl, void *user, _uint8 rid, _uint16 rtype, void *rdata, _uint16 rlen)
{
	return ENOTSUP;
}


/******************************************************************************

	Interrupt out callback.

******************************************************************************/
void usb_int_out_cbf (struct usbd_urb *urb, struct usbd_pipe *pipe, void *hdl)
{
    pipe = pipe, hdl = hdl;

    pthread_sleepon_lock ();
    pthread_sleepon_signal (urb);
    pthread_sleepon_unlock ();

    return;
}


/******************************************************************************

  Removal callback, cleans up the driver.

******************************************************************************/
int usb_int_out_wait( struct usbd_urb *urb, struct usbd_pipe *pipe )
{
	_uint32     ustatus;
	_uint32     alen;

	pthread_sleepon_lock ();
	while (usbd_urb_status (urb, &ustatus, &alen) == EBUSY) {
		pthread_sleepon_wait (urb);
	}

	pthread_sleepon_unlock ();

	usbd_urb_status (urb, &ustatus, &alen);

	// check for a stall condition and clear it
	if ((ustatus & USBD_USB_STATUS_MASK) == USBD_STATUS_STALL) {
		usbd_reset_pipe (pipe);
	}

	return ((ustatus & USBD_URB_STATUS_MASK) == USBD_STATUS_CMP ? EOK : EIO);
}


/******************************************************************************

	Client is sending output/feature report to device

******************************************************************************/
int usb_report_write( int reg_hdl, void *user, _uint8 rid, _uint16 rtype, void *rdata, _uint16 rlen  )
{
	usb_t                  *usb = user;
	egalax_data_packet_t   *data = (egalax_data_packet_t *) rdata;

	switch (data->cmd) {
	case SET_RAW_DATA:
		/* Make a connection back to calib to send a pulse for the next data point */
		if (usb->coid == NULL) {
			if ((usb->coid = name_open ("calib", 0)) == -1) {
				hid_slogf (_SLOGC_USB, 3, 0, "devh-egalax.so - name_open failed: %d", errno);
				break;
			}
		}

		/* Send the raw data to the controller as it comes in */
		egalax_send_calib_data (usb, data);
		break;
	case SET_REPORT_MODE:
		/* Set the report mode to RAW, as calib is waiting for raw points */
		egalax_set_mode (usb, RAW_MODE);
		break;
	case CALCULATE_CALIB_PARAMETERS_DATA:
		/* Tell the controller to calculate the calibration parameters based on the data we have just sent it */
		egalax_calculate_calib_params (usb);

		/* Set the report mode back to Calibrated */
		egalax_set_mode (usb, CALIBRATED_MODE);
		break;
	default:
		fprintf (stderr, "devh-egalax.so: Received an unknown command from somewhere\n");
		break;
	}

	return (EOK);
}


/******************************************************************************

	Fetch the idle time.

******************************************************************************/
int usb_get_idle (int reg_hdl, void *user, _uint8 rid, _uint16 *idle_rate)
{
	*idle_rate = 0;

	return (EOK);
}


/******************************************************************************

	Set the idle time.

******************************************************************************/
int usb_set_idle (int reg_hdl, void *user, _uint8 rid, _uint16 idle_rate)
{
	return (EOK);
}


/******************************************************************************

	Get the protocol.

******************************************************************************/
int usb_get_protocol (int reg_hdl, void *user, _uint8 *protocol)
{
	return (EOK);
}


/******************************************************************************

	Set the protocol.

******************************************************************************/
int usb_set_protocol (int reg_hdl, void *user, _uint8 protocol)
{
	return (EOK);
}


/******************************************************************************


******************************************************************************/
int usb_get_indexed_string (int reg_hdl, void *user, _uint16 index, void **string)
{
	usb_t *usb = user;

	*string = usbd_string (usb->device, index, 0); 

	return (*string ? EOK : ENOENT);
}


/******************************************************************************


******************************************************************************/
int usb_get_string (int reg_hdl, void *user, _uint16 request, void **string)
{
	usb_t *usb = user;
	usbd_device_descriptor_t *device_desc;

	if ((device_desc = usbd_device_descriptor( usb->device, NULL)) == NULL)
		return (EIO);

	switch (request) {
	case HIDD_STRING_MANUFACTURER:
		*string = usbd_string (usb->device, device_desc->iManufacturer, 0); // Lang not supported
		break;
	case HIDD_STRING_PRODUCT:
		*string = usbd_string (usb->device, device_desc->iProduct, 0);
		break;
	case HIDD_STRING_SERIAL_NUMBER:
		*string = usbd_string (usb->device, device_desc->iSerialNumber, 0);
		break;
	default:
		return (ENOENT);
	}

	return (*string ? EOK : ENOENT);
}


/******************************************************************************


******************************************************************************/
int usb_reset (int reg_hdl, void *user)
{
	usb_t *usb = user;

	return (usb_device_reset (usb));
}


/******************************************************************************


******************************************************************************/
int usb_rbuffer_alloc (_uint16 size, void **bptr)
{

	if ((*bptr = usbd_alloc (size)) == NULL) {
			return (ENOMEM);
	}
	
	memset (*bptr, 0xff, size);
	return (EOK);
}


/******************************************************************************


******************************************************************************/
int usb_rbuffer_free (void *bptr)
{
	usbd_free (bptr);
	return (EOK);
}


/******************************************************************************


******************************************************************************/
int usb_stop (void *dll_hdl)
{
	usb_t	*usb_device, *next = NULL;
	int		x;

	// driver is stopping detach from all usb devices
	for (usb_device = TAILQ_FIRST (&USBCtrl.dlist); usb_device; usb_device = next) {
		next = TAILQ_NEXT (usb_device, dlink);
		usb_free (usb_device);
	}

	// disconnect from USB stack
	for (x = 0; x < USBCtrl.num_conn; x++) {
		usbd_disconnect (USBCtrl.connection[x]);
	}

	if (!USBCtrl.num_conn) {	// stuck in init function; destroy mem alloced in usbd_connect()
		usbd_disconnect (NULL);
	}
		
	return (EOK);
}


/******************************************************************************


******************************************************************************/
int usb_start_driver (char *options)
{
	usbd_funcs_t funcs = {_USBDI_NFUNCS, usb_insertion, usb_removal, NULL};
	usbd_device_ident_t interest = {USB_VENDOR, USBD_CONNECT_WILDCARD, USBD_CONNECT_WILDCARD, USBD_CONNECT_WILDCARD, USBD_CONNECT_WILDCARD};
	usbd_connect_parm_t parm = {NULL, USB_VERSION, USBD_VERSION, 0, 0, 0, 0, 0, 0, USBD_CONNECT_WAIT};

	int status = EOK;
	char *value;

	USBCtrl.num_conn = 0;
	USBCtrl.verbose = 0;
	USBCtrl.cflags = 0;
	USBCtrl.sensitivity = 128;
	TAILQ_INIT (&USBCtrl.dlist);

	parm.funcs = &funcs;
	parm.ident = &interest;

	while (options && *options != '\0') {
		switch (getsubopt (&options, usb_opts, &value)) {
		case WAIT_USB:
			if (value != NULL) {
				parm.connect_wait = strtol (value, 0, 10);
			}
			break;
		case USB_SERVER:
			parm.path = value;
			break;
		case VENDOR_ID:
			if (value != NULL) {
				interest.vendor = strtoul (value, 0, 0);
			}
			break;
		case DEVICE_ID:
			if (value != NULL) {
				interest.device = strtoul (value, 0, 0);
			}
			break;
		case VERBOSE:
			USBCtrl.verbose = strtol (value, 0, 10);
			break;
		case NO_INIT:
			USBCtrl.cflags |= FLAG_NO_INIT;
			break;
		case GET_CONTROLLER_INFO:
			USBCtrl.cflags |= FLAG_GET_CONTROLLER_INFO;
			break;
		case HID:
			hid_slogf (_SLOGC_USB, 1, 0, "devh-egalax.so - HID command line option has been deprecated.  Please refer to the documentation for devh-egalax.so.");
			break;
		case SENSITIVITY:
			USBCtrl.sensitivity = strtol (value, 0, 10);
			break;
		case DELAY:
			USBCtrl.delay = strtol (value, 0, 10);
			break;
		default:
			fprintf (stderr, "Invalid option");
			break;
		}
	}

	// connect to stack
	if ((status = usbd_connect (&parm, &USBCtrl.connection[USBCtrl.num_conn])) != EOK) {
		hid_slogf (_SLOGC_USB, 3, 0, "HID Class/init - %s", strerror (status));
	}
	else {
		USBCtrl.num_conn++;
	}

	return (status);
}


/******************************************************************************

 register driver with hid class 

******************************************************************************/
int usb_init (void *dll_hdl, dispatch_t *dpp, io_hid_self_t *ioh, char *options)
{
	// register driver with io-hid
	USBCtrl.dll_hdl = dll_hdl;
	USBCtrl.ioh = ioh;

	if (usb_start_driver (options) != EOK) {
		hid_slogf (_SLOGC_USB, 2, 0, "error starting USB HID class driver");
	}

	return (0);
}
